/*
 * Daniel Henninger <daniel@ncsu.edu>
 *
 * Source and Documentation can be found at:
 * http://stonecold.unity.ncsu.edu/software/mod_auth_kerb/
 *
 * $Id: mod_auth_kerb.c,v 4.11 2002/04/09 19:27:06 daniel Exp $
 *
 * Originally by James E. Robinson, III <james@ncstate.net>
 *
 */

/*
 * Further modifications to this code done at the National Center
 * for Supercomputing Applications by Von Welch <vwelch@ncsa.edu>
 *
 * And also Modifications for sending back WWW-Authenticate headers on
 * each 401 by Mark Mentovai <mark@moxienet.com>
 *
 * KrbDefaultInstance and KrbAnyInstance directives added and
 * some inconsistencies fixed by David MacKenzie <djm@uu.net>.
 *
 * KrbAuthFailStatus directive added by
 * Neil Gulati <anilg@geocities.com> <ngulati@scu.edu.au>.
 * Using 403 convinces browser not to ask user for password.
 *
 * earlf - Added Authoritative control to allow passing on to lower
 *         modules if and only if the user-id is not known to this
 *         module. A known user with a faulty or absent password still
 *         causes an AuthRequired. The default is 'Authoritative', i.e.
 *         no control is passed along.
 */

/* ====================================================================
 * Copyright (c) 1996-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * Portions developed at the National Center for Supercomputing
 * Applications at the University of Illinois at Urbana-Champaign.
 */

static char rcsid[] = "$Id: mod_auth_kerb.c,v 4.11 2002/04/09 19:27:06 daniel Exp $";

/* Ouch, conflicting header file.  Make sure ssl des.h is ignored. */
#ifdef APACHE_SSL
#undef APACHE_SSL
#endif /* APACHE_SSL */

#ifdef STRONGHOLD
#undef STRONGHOLD
#endif /* STRONGHOLD */

#ifdef APXS2
#include "apr_strings.h"
#include "apr_lib.h"
#include "ap_config.h"
#endif /* APXS2 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

/*
 * Prevent warning about closesocket redefinition (Apache's
 * ap_config.h and Kerberos' cc-unix.h both define it as close)
 */
#ifdef closesocket
#undef closesocket
#endif /* closesocket */

#ifdef HESIOD
#include <hesiod.h>
#endif /* HESIOD */

#if (defined(DUAL_AUTH)) && (! defined(KRB5))
#define KRB5
#endif

#if (defined(DUAL_AUTH)) && (! defined(KRB4))
#define KRB4
#endif

#ifdef KRB5
#include "krb5.h"
#endif

#ifdef KRB4
#include "des.h"
#include "krb.h"
#endif

#ifdef KRB5
#define KRB5_DEFAULT_OPTIONS 0

#ifndef KRB5_SAVE_CREDENTIALS
#define KRB5_DEFAULT_LIFE    60 * 5 /* 5 minutes */
#else
#define KRB5_DEFAULT_LIFE    60 * 30 /* 30 minutes */
#endif

krb5_data tgtname = {
   0,
   KRB5_TGS_NAME_SIZE,
   KRB5_TGS_NAME
};

#ifdef KRB5_VERIFY_TICKET

#ifndef KRB5_DEFAULT_AUTH_SERVICE
#define KRB5_DEFAULT_AUTH_SERVICE	"www"
#endif /* !KRB5_DEFAULT_AUTH_SERVICE */

#endif /* KRB5_VERIFY_TICKET */

#endif

#ifdef KRB5_SAVE_CREDENTIALS
/* Are we saving our credentials? */
static int krb5_save_credentials;
#endif

#ifdef KRB4_SAVE_TICKETS
/* Are we saving our tickets? */
static int krb4_save_tickets;
#endif

#ifndef MAX_KDATA_LEN
#define MAX_KDATA_LEN 2048  /* Biggest Kerberos passwd string */
#endif

/* make this the default (v4) */
#ifdef KRB_DEF_REALM
#ifdef KRB_REALM
#undef KRB_REALM
#endif /* KRB_REALM */
#define KRB_REALM      KRB_DEF_REALM
#endif /* KRB_DEF_REALM */

/* default location for srvtab file */
#ifndef KRB_V4_SRVTAB
#define KRB_V4_SRVTAB  KEYFILE
#endif

#ifdef DEFAULT_TKT_LIFE
#undef DEFAULT_TKT_LIFE
#define DEFAULT_TKT_LIFE 1  /* default tkt life to 5 mins */
#endif

#ifdef KRB4
/* our httpd server kerberos realm */
static char realm[REALM_SZ];
#endif

#ifdef APXS2
#define ap_table_get apr_table_get
#define ap_table_set apr_table_set
#define ap_palloc apr_palloc
#define ap_pcalloc apr_pcalloc
#define ap_pstrcat apr_pstrcat
#define ap_log_reason(a,b,c) ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, APR_SUCCESS, c, a)
#endif /* APXS2 */

/* forward declaration; see definition at end of file */
#ifdef APXS2
module AP_MODULE_DECLARE_DATA kerb_auth_module;
#else
module kerb_auth_module;
#endif /* APXS2 */

/* forward declaration for KrbAuthFailStatus function, see near end of file */
#ifdef APXS2
static const char *set_fail_status_slot(cmd_parms *cmd, void *offset,
				const char *arg);
#else
static const char *set_fail_status_slot(cmd_parms *cmd, char *struct_ptr,
                                        char *arg);
#endif /* APXS2 */

/* configuration record definition */
typedef struct  {
   char *krb_auth_realm;
   char *krb_lifetime;
   char *krb_default_instance;
   int krb_any_instance;
#if defined(KRB5_SAVE_CREDENTIALS) || defined(KRB4_SAVE_TICKETS)
   int krb_expire_reauth;
#endif /* defined(KRB5_SAVE_...... */
#ifdef KRB4
   char *krb4_srvtab;
#ifdef KRB4_SAVE_TICKETS
   int krb4_save_tickets;
#endif /* KRB4_SAVE_TICKETS */
#endif /* KRB4 */
#ifdef KRB5
   char *krb5_keytab;
#ifdef KRB5_SAVE_CREDENTIALS
   int krb5_save_credentials;
   int krb5_forwardable;
#endif /* KRB5_SAVE_CREDENTIALS */
#endif /* KRB5 */
   int krb_auth_fail_status; /* KrbAuthFailStatus=401 or 403 */
   int krb_auth_authoritative;
} kerb_auth_config_rec;

/* map the valid commands to the config record */
#ifdef APXS2
static const command_rec kerb_auth_cmds[] =
{
	AP_INIT_TAKE1("KrbAuthRealm", ap_set_string_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb_auth_realm),
		OR_AUTHCFG, "Kerberos realm in which to authenticate users."),

	AP_INIT_TAKE1("KrbLifetime", ap_set_string_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb_lifetime),
		OR_AUTHCFG, "Kerberos ticket lifetime."),

	AP_INIT_TAKE1("KrbDefaultInstance", ap_set_string_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb_default_instance),
		OR_AUTHCFG, "Kerberos instance for users who don't specify one; defaults to empty."),

	AP_INIT_FLAG("KrbAnyInstance", ap_set_flag_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb_any_instance),
		OR_AUTHCFG, "Allow users to specify an instance?"),

#if defined(KRB5_SAVE_CREDENTIALS) || defined(KRB4_SAVE_TICKETS)
	AP_INIT_FLAG("KrbExpireReauth", ap_set_flag_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb_expire_reauth),
		OR_AUTHCFG, "Reauth if tickets expire?"),
#endif /* defined(KRB5_SAVE_...... */

#ifdef KRB4
	AP_INIT_TAKE1("Krb4Srvtab", ap_set_file_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb4_srvtab),
		OR_AUTHCFG, "Kerberos v4 srvtab."),

#ifdef KRB4_SAVE_TICKETS
	AP_INIT_FLAG("Krb4SaveTickets", ap_set_flag_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb4_save_tickets),
		OR_AUTHCFG, "Save the v4 ticket file?"),
#endif /* KRB4_SAVE_TICKETS */
#endif /* KRB4 */

#ifdef KRB5
	AP_INIT_TAKE1("Krb5Keytab", ap_set_file_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb5_keytab),
		OR_AUTHCFG, "Kerberos v5 keytab."),

#ifdef KRB5_SAVE_CREDENTIALS
	AP_INIT_FLAG("Krb5SaveCredentials", ap_set_flag_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb5_save_credentials),
		OR_AUTHCFG, "Save the v5 credential cache?"),

	AP_INIT_FLAG("Krb5Forwardable", ap_set_flag_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb5_forwardable),
		OR_AUTHCFG, "Set tickets to be forwardable?"),
#endif /* KRB5_SAVE_CREDENTIALS */
#endif /* KRB5 */

	AP_INIT_TAKE1("KrbAuthFailStatus", set_fail_status_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb_auth_fail_status),
		OR_AUTHCFG, "HTTP status code to return if authentication fails."),

	AP_INIT_FLAG("KrbAuthAuthoritative", ap_set_flag_slot,
		(void *) APR_XtOffsetOf(kerb_auth_config_rec, krb_auth_authoritative),
		OR_AUTHCFG, "Set to 'off' to allow access control to be passed along to lower modules if the UserID is not known to this module?"),

	{ NULL }
};
#else
command_rec kerb_auth_cmds[] = {
   { "KrbAuthRealm", ap_set_string_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb_auth_realm),
    OR_AUTHCFG, TAKE1, "Kerberos realm in which to authenticate users." },

   { "KrbLifetime", ap_set_string_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb_lifetime),
    OR_AUTHCFG, TAKE1, "Kerberos ticket lifetime." },

   { "KrbDefaultInstance", ap_set_string_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb_default_instance),
    OR_AUTHCFG, TAKE1, "Kerberos instance for users who don't specify one; defaults to empty." },

   { "KrbAnyInstance", ap_set_flag_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb_any_instance),
    OR_AUTHCFG, FLAG, "Allow users to specify an instance?" },

#if defined(KRB5_SAVE_CREDENTIALS) || defined(KRB4_SAVE_TICKETS)
   { "KrbExpireReauth", ap_set_flag_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb_expire_reauth),
    OR_AUTHCFG, FLAG, "Reauth if tickets expire?" },
#endif /* defined(KRB5_SAVE_...... */

#ifdef KRB4
   { "Krb4Srvtab", ap_set_file_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb4_srvtab),
    OR_AUTHCFG, TAKE1, "Kerberos v4 srvtab." },

#ifdef KRB4_SAVE_TICKETS
   { "Krb4SaveTickets", ap_set_flag_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb4_save_tickets),
    OR_AUTHCFG, FLAG, "Save the v4 ticket file?" },
#endif /* KRB4_SAVE_TICKETS */
#endif /* KRB4 */

#ifdef KRB5
   { "Krb5Keytab", ap_set_file_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb5_keytab),
    OR_AUTHCFG, TAKE1, "Kerberos v5 keytab." },

#ifdef KRB5_SAVE_CREDENTIALS
   { "Krb5SaveCredentials", ap_set_flag_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb5_save_credentials),
    OR_AUTHCFG, FLAG, "Save the v5 credential cache?" },

   { "Krb5Forwardable", ap_set_flag_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb5_forwardable),
    OR_AUTHCFG, FLAG, "Set tickets to be forwardable?" },
#endif /* KRB5_SAVE_CREDENTIALS */
#endif /* KRB5 */

   /* KrbAuthFailStatus directive to store HTTP auth fail status code */
   { "KrbAuthFailStatus", set_fail_status_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb_auth_fail_status),
    OR_AUTHCFG, TAKE1, "HTTP status code to return if authentication fails." },

   { "KrbAuthAuthoritative", ap_set_flag_slot,
    (void*)XtOffsetOf(kerb_auth_config_rec, krb_auth_authoritative),
    OR_AUTHCFG, FLAG, "Set to 'off' to allow access control to be passed along
to lower modules if the UserID is not known to this module?" },
   { NULL }
};
#endif /* APXS2 */

/* Tables for converting binary values to and from hexadecimal */

static char hex[] = "0123456789abcdef";

static char dec[256] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /*   0 -  15 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /*  16 -  37 */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* ' ' - '/' */
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 0, 0, 0, 0, 0,   /* '0' - '?' */
    0,10,11,12,13,14,15, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* '@' - 'O' */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 'P' - '_' */
    0,10,11,12,13,14,15, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* '`' - 'o' */
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   /* 'p' - DEL */
};

/*************************************************************************
 * kdata_to_str -- convert 8-bit char array to ascii string
 *
 * Accepts:  input array and length
 * Returns:  a pointer to the result, or null pointer on malloc failure
 *           The caller is responsible for freeing the returned value.
 *
 ************************************************************************/
static char *kdata_to_str(char *in_data, int length)
{
   char *result, *p;
   int i;
 
   p = result = malloc(length*2+1);

   if (!result) {
      return (char *) NULL;
   }

   for (i=0; i < length; i++) {
      *p++ = hex[(in_data[i]>>4)&0xf];
      *p++ = hex[(in_data[i])&0xf];
   }

   *p++ = '\0';

   return result;
}
 
 
/*************************************************************************
 * str_to_kdata -- Converts ascii string to a (binary) char array
 *
 * Accepts: string to convert
 *          pointer to output array
 * Returns: length of output array, NULL on failure
 *
 ************************************************************************/
static int str_to_kdata(const char *in_str, char *out_str)
{
   int inlen, outlen;
 
   inlen = strlen(in_str);
   if (inlen & 1) {
      return -1;  /* must be even number, in this scheme */
   }

   inlen /= 2;

   if (inlen > MAX_KDATA_LEN) {
      return -1;
   }
 
   for (outlen=0; *in_str; outlen++, in_str += 2) {
      out_str[outlen] = (dec[in_str[0]]<<4) + dec[in_str[1]];
   }

   return outlen;
}

/*************************************************************************
 * kerb_get_server_realm -- retrieve the realm information for the server
 *
 * Accepts:  request record
 * Returns:  sets the global variable 'realm' to the server's realm
 *
 ************************************************************************/
#ifdef KRB4
static void kerb_get_server_realm( request_rec *r)
{
   /* get the config information */
   kerb_auth_config_rec *sec =
      (kerb_auth_config_rec *)ap_get_module_config(r->per_dir_config,
                              &kerb_auth_module);

   /*
    * did the server config specify a different realm than compiled in default?
    */
   if(sec->krb_auth_realm) {
      strncpy(realm, sec->krb_auth_realm, REALM_SZ);
   } else {
      /* if not, try and get the one from the system */
      if (krb_get_lrealm(realm, 1) != KSUCCESS) {
         /* else, resort to the compiled in default */
         strncpy(realm, KRB_REALM, REALM_SZ);
      }
   }
}
#endif

/*************************************************************************
 * kerb_validate_user_pass -- validate user with principal and passwd
 *
 * Accepts:  request record and sent password
 * Returns:  OK, DECLINED, HTTP_UNAUTHORIZED or SERVER_ERROR
 * KrbAuthFailStatus: now returns AUTH_FORBIDDEN if configured to do so
 *
 ************************************************************************/
#ifdef KRB4
static int kerb_validate_user_pass_V4( request_rec *r, const char *sent_pw )
{
   kerb_auth_config_rec *sec =
      (kerb_auth_config_rec *)ap_get_module_config(r->per_dir_config,
                              &kerb_auth_module);
   char errstr[MAX_STRING_LEN];
   int krbval;
   int kpass_ok = 0;   /* not allowed until we say so */
   int lifetime = DEFAULT_TKT_LIFE;
   char krb_tkt_file[MAXPATHLEN - 1];
   char krb_tmp_tkt_file[MAXPATHLEN - 1];
   CREDENTIALS c;
   char pname[ANAME_SZ];
   char pinst[INST_SZ];
   char prealm[REALM_SZ];
   int code;
   char *instance;
   char *username;
   char *cn;
   int removetf;

   if (sec->krb_lifetime) {
      lifetime = atoi(sec->krb_lifetime);
   }

   /* need to be sure ticket file is set to something reasonable
    * since multiple processes could be authenticating at once
    * better make sure files don't conflict
    */
   sprintf(krb_tkt_file, "/tmp/apache_tkt_%ld", (long)getpid());
   krb_set_tkt_string(krb_tkt_file);

   /* Default the instance. */
#ifdef APXS2
   username = strdup(r->user);
#else
   username = strdup(r->connection->user);
#endif /* APXS2 */
   instance = strchr(username, '.');
   if (instance) {
     if (!sec->krb_any_instance) {
       sprintf(errstr, "specifying Kerberos instance is denied by server configuration");
       ap_log_reason (errstr, r->uri, r);
       return sec->krb_auth_fail_status;
     } else {
       *instance++ = '\0';
     }
   } else if (sec->krb_default_instance) {
     instance = sec->krb_default_instance;
   } else {
     instance = "";
   }

#ifdef KRB4_SAVE_TICKETS
   if (! sec->krb4_save_tickets) {
      removetf = 1;
   }
   else {
      removetf = 0;
   }
#else
   removetf = 1;
#endif /* KRB4_SAVE_TICKETS */

#ifdef KRB4_SAVE_TICKETS
   /*
   * krb4_save_tickets is set in kerb_authenticate_user
   */
   if (sec->krb4_save_tickets || krb4_save_tickets) {
      /*
       * Save the tickets somewhere a CGI script can find them:
       * /tmp/apache_tkt_<username>.  To make it a valid file name
       * we change any '/' into '.' in <username>.
       */
#ifdef APXS2
      sprintf(krb_tmp_tkt_file, "/tmp/apache_tkt_%s", r->user);
#else
      sprintf(krb_tmp_tkt_file, "/tmp/apache_tkt_%s", r->connection->user);
#endif /* APXS2 */
      for (cn = krb_tmp_tkt_file + 16; *cn; cn++)
	if (*cn == '/')
	  *cn = '.';
      krb_set_tkt_string(krb_tmp_tkt_file);

      if (!access(krb_tmp_tkt_file, F_OK) && sec->krb_expire_reauth) {
         if ((code = krb_get_tf_realm(krb_tmp_tkt_file, prealm)) != KSUCCESS) {
            sprintf(errstr,
               "krb_tkt_file(): Unable to get realm of ticket file - %d", code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if ((code = tf_init(krb_tmp_tkt_file, R_TKT_FIL)) != KSUCCESS) {
            sprintf(errstr,
               "krb_tkt_file(): Unable to open ticket file - %d", code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if ((code = tf_get_pname(pname)) != KSUCCESS) {
            sprintf(errstr,
               "krb_tkt_file(): Unable to get principal name - %d", code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if ((code = tf_get_pinst(pinst)) != KSUCCESS) {
            sprintf(errstr,
               "krb_tkt_file(): Unable to get principal instance - %d", code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }
            
         if ((code = tf_get_cred(&c)) != KSUCCESS) {
            sprintf(errstr,
               "krb_tkt_file(): Unable to retrieve ticket - %d", code);
            ap_log_reason (errstr, r->uri, r);
            tf_close();
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if (time(0) > (c.issue_date + (((unsigned char) c.lifetime)*5*60))) {
            tf_close();
            dest_tkt();
            if(sec->krb_auth_fail_status == HTTP_UNAUTHORIZED)
                ap_table_set (r->err_headers_out, "WWW-Authenticate",
						"KerberosV4");
            return sec->krb_auth_fail_status;
         }

         tf_close();

         removetf = 1;
      }
      else {
         strcpy(krb_tkt_file, krb_tmp_tkt_file);
      }
   }

   krb_set_tkt_string(krb_tkt_file);
#endif /* KRB4_SAVE_TICKETS */

   /* see if they are who they say they are */
   krbval = krb_get_pw_in_tkt(username, instance, realm,
      "krbtgt", realm, lifetime, (char *)sent_pw);
 
   /* destroy all the evidence */
   /* well, it is still in memory on the browser.. */
   if (removetf) {
      dest_tkt();
   }

   /* see what kerberos told us */
   switch (krbval) {
      case INTK_OK:
      case INTK_W_NOTALL:
         kpass_ok = 1;
         break;   
 
       default:
         sprintf(errstr, "krb_get_pw_in_tkt(%s.%s): %s",
#ifdef APXS2
		 r->user, instance, krb_err_txt[krbval]);
#else
		 r->connection->user, instance, krb_err_txt[krbval]);
#endif /* APXS2 */

         /* first, the error log */
         ap_log_reason (errstr, r->uri, r);
         break;
   }

   /* are they ok? */
   if ( kpass_ok ) {
      return OK;
   }

   /* if not, let somebody else take care of it */
   return sec->krb_auth_fail_status;
}
#endif

#ifdef KRB5
/*
 * Did the server config specify a default instance and/or the realm?
 * Sets *meep and *clientp, and returns OK, or an error code.
 * KrbAuthFailStatus: returns krb_auth_fail_status instead of HTTP_UNAUTHORIZED
 */
static int kerb_default_user(request_rec *r, krb5_context kcontext, krb5_principal *meep,
			     char **clientp)
{
   kerb_auth_config_rec *sec =
      (kerb_auth_config_rec *)ap_get_module_config(r->per_dir_config,
                              &kerb_auth_module);
   krb5_error_code  code;
   char errstr[MAX_STRING_LEN], *newprinc, *to, *from;
   size_t newlen;
   int set_instance = 0, set_realm = 0;
   
#ifdef APXS2
   if (strchr(r->user,'/')) {
#else
   if (strchr(r->connection->user,'/')) {
#endif /* APXS2 */
     if (!sec->krb_any_instance) {
       sprintf(errstr, "specifying Kerberos instance is denied by server configuration");
       ap_log_reason (errstr, r->uri, r);
       return sec->krb_auth_fail_status;
     }
   } else if (sec->krb_default_instance) {
     set_instance = 1;
   }

   if (sec->krb_auth_realm) {
     set_realm = 1;
   }

#ifdef APXS2
   newlen = strlen(r->user) + 1;
#else
   newlen = strlen(r->connection->user) + 1;
#endif /* APXS2 */
   if (set_instance)
     newlen += strlen(sec->krb_default_instance) + 1;
   if (set_realm)
     newlen += strlen(sec->krb_auth_realm) + 1;
   to = newprinc = (char *)ap_palloc(r->pool, newlen);
   
   /* Copy the user name and any /instances.  */
#ifdef APXS2
   for (from = r->user; *from;)
#else
   for (from = r->connection->user; *from;)
#endif /* APXS2 */
     *to++ = *from++;

   if (set_instance) {
     *to++ = '/';
     for (from = sec->krb_default_instance; *from;)
       *to++ = *from++;
   }

   if (set_realm) {
     *to++ = '@';
     for (from = sec->krb_auth_realm; *from;)
       *to++ = *from++;
   }

   *to++ = '\0';

   /* Break the revised principal name into its parts.  */
   if (code = krb5_parse_name(kcontext, newprinc, meep)) {
      sprintf(errstr, "krb5_parse_name(): Parsing name %s - %s (%d)",
	      newprinc, error_message(code), code);
      ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
      return HTTP_INTERNAL_SERVER_ERROR;
#else
      return SERVER_ERROR;
#endif /* APXS2 */
   }

   /* Canonicalize the principal name.  */
   if (code = krb5_unparse_name(kcontext, *meep, clientp)) {
      sprintf(errstr, "krb5_unparse_name(): Unparsing name - %s (%d)",
	      error_message(code), code);
      ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
      return HTTP_INTERNAL_SERVER_ERROR;
#else
      return SERVER_ERROR;
#endif /* APXS2 */
   }

   return OK;
}

static int kerb_validate_user_pass_V5( request_rec *r, const char *sent_pw )
{
   kerb_auth_config_rec *sec =
      (kerb_auth_config_rec *)ap_get_module_config(r->per_dir_config,
                              &kerb_auth_module);
   krb5_context     kcontext;
   krb5_ccache      ccache = NULL;
   krb5_principal   me;
   krb5_principal   server;
   krb5_creds       my_creds, creds;
   krb5_timestamp   now;
   krb5_cc_cursor   cur;
   long             lifetime = KRB5_DEFAULT_LIFE;
   int              options = KRB5_DEFAULT_OPTIONS;
   krb5_preauthtype *preauth = NULL;
   krb5_error_code  code;
   int		    userok;
#ifdef KRB5_VERIFY_TICKET
   /* Service to use for authentication */
   char             *auth_service = KRB5_DEFAULT_AUTH_SERVICE;
   krb5_ccache	    ccache_save;
#ifdef KRB5_SAVE_CREDENTIALS
   /* Do we need to get a new ticket-granting-ticket? */
   int              get_tgt = 0;
#endif /* KRB5_SAVE_CREDENTIALS */
#endif /* KRB5_VERIFY_TICKET */


   char             *client_name;
   char             errstr[MAX_STRING_LEN];

   if (sec->krb_lifetime) {
      lifetime = atoi(sec->krb_lifetime);
   }

   krb5_init_context(&kcontext);
   krb5_init_ets(kcontext);

   userok = kerb_default_user(r, kcontext, &me, &client_name);
   if (userok != OK)
     return userok;

   if (code = krb5_timeofday(kcontext, &now)) {
      sprintf(errstr, "krb5_timeofday(): Getting time of day - error %s (%d)",
           error_message(code), code);
      ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
      return HTTP_INTERNAL_SERVER_ERROR;
#else
      return SERVER_ERROR;
#endif /* APXS2 */
   }

#ifdef KRB5_SAVE_CREDENTIALS
   /*
   * krb5_save_credentials is set in kerb_authenticate_user
   */
   if (sec->krb5_save_credentials || krb5_save_credentials) {
      /*
       * Save the credentials somewhere a CGI script can find them:
       * /tmp/krb5cc_<username>.  To make it a valid file name
       * we change any '/' into '.' in <username>.
       */
      char            cache_name[MAX_STRING_LEN], *cn;

#ifdef APXS2
      sprintf(cache_name, "/tmp/krb5cc_%s", r->user);
#else
      sprintf(cache_name, "/tmp/krb5cc_%s", r->connection->user);
#endif /* APXS2 */
      for (cn = cache_name + 12; *cn; cn++)
	if (*cn == '/')
	  *cn = '.';

      if (code = krb5_cc_resolve(kcontext, cache_name, &ccache)) {
         sprintf(errstr,
               "krb5_cc_resolve(%s): Getting credential cache - %s (%d)",
               cache_name, error_message(code), code);
         ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
         return HTTP_INTERNAL_SERVER_ERROR;
#else
         return SERVER_ERROR;
#endif /* APXS2 */
      }

      if (!access(cache_name, F_OK) && sec->krb_expire_reauth) {
         if ((code = krb5_cc_set_flags(kcontext, ccache, 0))) {
            sprintf(errstr,
               "krb5_cc_set_flags(): Unable to set flags on cache - %s (%d)",
               error_message(code), code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if ((code = krb5_cc_start_seq_get(kcontext, ccache, &cur))) {
            sprintf(errstr,
               "krb5_cc_start_seq_get(): Unable to start retrieving tickets - %s (%d)",
               error_message(code), code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if ((code = krb5_cc_next_cred(kcontext, ccache, &cur, &creds))) {
            /* No tickets in cache, make them auth! */
            krb5_cc_destroy(kcontext, ccache);
            if(sec->krb_auth_fail_status == HTTP_UNAUTHORIZED)
                ap_table_set (r->err_headers_out, "WWW-Authenticate",
					"KerberosV5");
            return sec->krb_auth_fail_status;
         }

         if ((code = krb5_cc_end_seq_get(kcontext, ccache, &cur))) {
            sprintf(errstr,
               "krb5_cc_start_seq_get(): Unable to finish retrieving tickets - %s (%d)",
               error_message(code), code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if ((code = krb5_cc_set_flags(kcontext, ccache, KRB5_TC_OPENCLOSE))) {
            sprintf(errstr,
               "krb5_cc_set_flags(): Unable to set flags on cache - %s (%d)",
               error_message(code), code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if (creds.times.endtime < now) {
            krb5_cc_destroy(kcontext, ccache);
            if(sec->krb_auth_fail_status == HTTP_UNAUTHORIZED)
                ap_table_set (r->err_headers_out, "WWW-Authenticate",
				"KerberosV5");
            return sec->krb_auth_fail_status;
         }

         ccache = NULL;
      }
      else {
         if (code = krb5_cc_initialize(kcontext, ccache, me)) {
            sprintf(errstr,
                  "krb5_cc_initialize(): Initializing credential cache - %s (%d)",
                  error_message(code), code);
            ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
            return HTTP_INTERNAL_SERVER_ERROR;
#else
            return SERVER_ERROR;
#endif /* APXS2 */
         }

         if (sec->krb5_forwardable) {
            options |= KDC_OPT_FORWARDABLE;
         }
	 
#ifdef KRB5_VERIFY_TICKET
	 /* Indicate that we need to get a ticket-granting-ticket */
	 get_tgt = 1;
#endif /* KRB5_VERIFY_TICKET */
      }
   }

   /*
   * If we're not saving the credentials, then we want to leave ccache == NULL
   */

#endif /* KRB5_SAVE_CREDENTIALS */

   memset((char *)&my_creds, 0, sizeof(my_creds));
    
   my_creds.client = me;

#ifdef KRB5_VERIFY_TICKET
   /* Get our service ticket so we can check it */
   code = krb5_sname_to_principal(kcontext,
				  NULL /* local hostname */,
				  auth_service,
				  KRB5_NT_SRV_HST,
				  &server);

   if (code) {
       sprintf(errstr,
	       "krb5_sname_to_principal(): Building server name for service %s - %s (%d)",
	       auth_service, error_message(code), code);
       ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
       return HTTP_INTERNAL_SERVER_ERROR;
#else
       return SERVER_ERROR;
#endif /* APXS2 */
   }

   /*
    * Save the cache and make sure we don't use a cache when doing
    * authentication. This way there is no chance of accidentally using
    * an existing cache to authenticate. This also keeps the authenticating
    * credentials from going into the cache in the first place.
    */
   ccache_save = ccache;
   ccache = NULL;

#else /* !KRB5_VERIFY_TICKET */
   /* Get tgt */
   if (code = krb5_build_principal_ext(kcontext, &server,
      krb5_princ_realm(kcontext, me)->length,
      krb5_princ_realm(kcontext, me)->data,
      tgtname.length, tgtname.data,
      krb5_princ_realm(kcontext, me)->length,
      krb5_princ_realm(kcontext, me)->data,
      0)) {
      sprintf(errstr, "krb5_build_principal_ext(): Building server name - %s (%d)", error_message(code), code);
      ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
      return HTTP_INTERNAL_SERVER_ERROR;
#else
      return SERVER_ERROR;
#endif /* APXS2 */
   }
#endif /* !KRB5_VERIFY_TICKET */

   my_creds.server = server;

   my_creds.times.starttime = 0;
   my_creds.times.endtime = now + lifetime;
   my_creds.times.renew_till = 0;

   code = krb5_get_in_tkt_with_password(kcontext,
      options, 0, NULL, preauth, sent_pw, ccache, &my_creds, 0);

#ifdef KRB5_VERIFY_TICKET
   /* Restore cache */
   ccache = ccache_save;

   if (!code) {
       char			*keytab_file = NULL;
       krb5_auth_context	auth_context = NULL;
       krb5_data		req_buffer;
       krb5_keytab		keytab = NULL;
       char			*keytab_name = NULL;


       /*
	* Check the credentials we received with the key we have in our
	* keytab and make sure that the credentials actually came from
	* the KDC and not someone faking a KDC response.
	*/

#ifdef KRB5_DEFAULT_KEYTAB
       keytab_name = KRB5_DEFAULT_KEYTAB;
#else  /** added to consider the .htaccess configuration.  - T.Huang **/
       if (sec->krb5_keytab) {
          keytab_name = sec->krb5_keytab;
       }
#ifdef KRB_V5_KEYTAB
       else {
          keytab_name = KRB_V5_KEYTAB;
       }
#endif /* KRB_V5_KEYTAB */
#endif /* KRB5_DEFAULT_KEYTAB */

       if (keytab_name) {
	   code = krb5_kt_resolve(kcontext, keytab_name, &keytab);

       } else {
	   code = krb5_kt_default(kcontext, &keytab);

       }

       if (code) {
	   sprintf(errstr,
		   "Resolving keytab file %s: %s (%d)",
		   (keytab_name ? keytab_name : "default"),
		   error_message(code), code);
	   ap_log_reason (errstr, r->uri, r);
	   goto verify_done;
       }

       
       /* Construct a ticket for our service with credentials */
       code = krb5_mk_req_extended(kcontext, &auth_context, 0 /* no options */,
				   NULL /* no input data */, &my_creds,
				   &req_buffer);

       if (code) {
	   sprintf(errstr,
		   "krb5_mk_req_extended(): %s (%d)",
		   error_message(code), code);
	   ap_log_reason (errstr, r->uri, r);
	   goto verify_done;
       }

       if (auth_context) {
	   /*
	    * Clear the auth context if it was set, otherwise krb5_rd_req()
	    * will fail.
	    */
	   krb5_auth_con_free(kcontext, auth_context);
	   auth_context = NULL;
       }

       /* Test the constructed ticket */
       code = krb5_rd_req(kcontext, &auth_context, &req_buffer, server,
			  keytab, NULL /* no options */,
			  NULL /* don't actually need ticket */);

       if (code) {
	   sprintf(errstr,
		   "krb5_rd_req(): %s (%d)",
		   error_message(code), code);
	   ap_log_reason (errstr, r->uri, r);
	   goto verify_done;
       }

   verify_done:	/* Jump here in case of error */
       
       if (keytab)
	   krb5_kt_close(kcontext, keytab);

       if (auth_context)
	   krb5_auth_con_free(kcontext, auth_context);

       if (req_buffer.data)
#ifdef KRB5LEGACY
           krb5_xfree(req_buffer.data);
#else
           krb5_free_data_contents(kcontext, &req_buffer);
#endif /* KRB5LEGACY */

       if (code) {
	   /* This frees the principals as well */
	   krb5_free_cred_contents(kcontext, &my_creds);

#ifdef APXS2
           return HTTP_INTERNAL_SERVER_ERROR;
#else
           return SERVER_ERROR;
#endif /* APXS2 */
       }

#ifdef KRB5_SAVE_CREDENTIALS	       
       /*
	* Now if we actually need to get a TGT, do it.
	*/
       if (get_tgt) {

	   /* Make sure my principal isn't freed */
	   me = my_creds.client;
	   my_creds.client = NULL;

	   krb5_free_cred_contents(kcontext, &my_creds);

	   memset((char *)&my_creds, 0, sizeof(my_creds));

	   my_creds.client = me;

	   if (code =
	       krb5_build_principal_ext(kcontext, &server,
					krb5_princ_realm(kcontext, me)->length,
					krb5_princ_realm(kcontext, me)->data,
					tgtname.length, tgtname.data,
					krb5_princ_realm(kcontext, me)->length,
					krb5_princ_realm(kcontext, me)->data,
					0)) {
	       sprintf(errstr,
		       "krb5_build_principal_ext(): Building server name for service %s - %s (%d)",
		       tgtname.data, error_message(code), code);
	       ap_log_reason (errstr, r->uri, r);
#ifdef APXS2
               return HTTP_INTERNAL_SERVER_ERROR;
#else
               return SERVER_ERROR;
#endif /* APXS2 */
	   }

	   my_creds.server = server;

	   my_creds.times.starttime = 0;
	   my_creds.times.endtime = now + lifetime;
	   my_creds.times.renew_till = 0;

	   code = krb5_get_in_tkt_with_password(kcontext, options, 0, NULL,
						preauth, sent_pw, ccache,
						&my_creds, 0);

	   if (code) {
	       sprintf(errstr,
		       "krb5_get_in_tkt_with_password: Error geting TGT after successful authentication");
	       ap_log_reason (errstr, r->uri, r);
	       /* Fall through and error will bt handled below */
	   }

       }
#endif /* KRB5_SAVE_CREDENTIALS */

   }
#endif /* KRB5_VERIFY_TICKET */

   /* This frees the principals as well */
   krb5_free_cred_contents(kcontext, &my_creds);

#ifdef KRB5_SAVE_CREDENTIALS
   if ((ccache != NULL) &&
          (sec->krb5_save_credentials || krb5_save_credentials)) {
      if (code == 0) {
         krb5_cc_close(kcontext, ccache);
      } else {
         krb5_cc_destroy(kcontext, ccache);
      }
   }
#endif /* KRB5_SAVE_CREDENTIALS */

   if (code) {
      switch(code) {
      case KRB5KRB_AP_ERR_BAD_INTEGRITY:
         sprintf(errstr,
               "krb5_get_in_tkt_with_password(): User \"%s\" - Password incorrect",
#ifdef APXS2
               r->user);
#else
               r->connection->user);
#endif /* APXS2 */
         break;

      case KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN:
         sprintf(errstr,
               "krb5_get_in_tkt_with_password(): Unknown user \"%s\"",
#ifdef APXS2
               r->user);
#else
               r->connection->user);
#endif /* APXS2 */
	 if (!(sec->krb_auth_authoritative)) {
	     /* fall through to a lower-level authentication type */
	    return DECLINED;
	 }
         break;

      default:
         sprintf(errstr,
               "krb5_get_in_tkt_with_password(): User \"%s\" - %s (%d)",
#ifdef APXS2
               r->user, error_message(code), code);
#else
               r->connection->user, error_message(code), code);
#endif /* APXS2 */
         break;
      }
      ap_log_reason (errstr, r->uri, r);

      return sec->krb_auth_fail_status;
   }

   return OK;
}
#endif

/*************************************************************************
 * kerb_validate_ticket -- validate the tkt sent by the browser
 *
 * Accepts:  request record
 * Returns:  OK, DECLINED, HTTP_UNAUTHORIZED or SERVER_ERROR
 *
 ************************************************************************/
#ifdef KRB4
static int kerb_validate_ticket_V4( request_rec *r )
{
   kerb_auth_config_rec *sec =
      (kerb_auth_config_rec *)ap_get_module_config(r->per_dir_config,
                              &kerb_auth_module);
   int krbval;
   char errstr[MAX_STRING_LEN];
   KTEXT_ST authent;
   char instance[INST_SZ];
   static AUTH_DAT kdata;
   char k4_srvtab[MAX_STRING_LEN];
   des_cblock session;                 /* Our session key */
   des_key_schedule schedule;          /* Schedule for our session key */
   char *type, *p;
   const char *sent_pw, *auth_line;

   /* get the auth line the user sent us */
   auth_line = ap_table_get (r->headers_in, "Authorization");

   /* did they send us the information required? */
   if ( auth_line == NULL ) {
      /* tell the client what to expect */
      if(sec->krb_auth_fail_status == HTTP_UNAUTHORIZED)
          ap_table_set (r->err_headers_out, "WWW-Authenticate", "KerberosV4");
      return sec->krb_auth_fail_status;
   }

   /* get the auth type sent to us */
   type = ap_getword(r->pool, &auth_line, ' ');

   /*
    * These two fields in the connection record are used elsewhere
    * in the Apache modules. (check_access)
    * They really appreciate them being set to useful values. :)
    */
#ifdef APXS2
   r->user = ap_getword(r->pool, &auth_line, ' ');
   r->ap_auth_type = "KerberosV4";
#else
   r->connection->user = ap_getword(r->pool, &auth_line, ' ');
   r->connection->ap_auth_type = "KerberosV4";
#endif /* APXS2 */

   sent_pw = ap_getword(r->pool, &auth_line, '\0');

   if ((authent.length = str_to_kdata(sent_pw, (char *)authent.dat)) == -1) {
      sprintf(errstr,"Invalid Kerberos authenticator");
      /* log it to the error file */
      ap_log_reason (errstr, r->uri, r);
      /* let somebody else take care of it */
      return sec->krb_auth_fail_status;
   }
 
   /* configure the srvtab file location */
   if (sec->krb4_srvtab) {
      strncpy(k4_srvtab, sec->krb4_srvtab, MAX_STRING_LEN);
   }
   else {
      strncpy(k4_srvtab, KRB_V4_SRVTAB, MAX_STRING_LEN);
   }

   /* configure the default instance */
   strcpy(instance, "*");

   /* Verify authenticator */
   if (k4_srvtab[0]) {
      krbval = krb_rd_req(&authent, "khttp", instance, 0L, &kdata, k4_srvtab);
   } else {
      /* not likely to work, but we'll try without the srvtab */
      krbval = krb_rd_req(&authent, "khttp", instance, 0L, &kdata, NULL);
   }
 
   /* see what the master says */
   /* this code should be similar to the other case statement */
   if (krbval) {
      /* No eh? log it so the admin can see what happened */
      sprintf(errstr, "krb_rd_req(): %s", krb_err_txt[krbval]);
      ap_log_reason (errstr, r->uri, r);
      /* let somebody else take care of it */
      return sec->krb_auth_fail_status;
   }
 
   /* Check Kerberos principal versus the username they sent us */
#ifdef APXS2
   if (strncmp(kdata.pname, r->user, ANAME_SZ)) {
#else
   if (strncmp(kdata.pname, r->connection->user, ANAME_SZ)) {
#endif /* APXS2 */
      sprintf(errstr, "Permission denied; name/username mismatch.");
      ap_log_reason (errstr, r->uri, r);
      /* let somebody else take care of it */
      return sec->krb_auth_fail_status;
   }

   /* Save the session key */
   memcpy(session, kdata.session, sizeof(des_cblock));
   key_sched(session, schedule);
   
   /* Construct the response for mutual authentication */
   authent.length = sizeof(des_cblock);
   memset(authent.dat, 0x00, sizeof(des_cblock));
   *((long *)authent.dat) = htonl(kdata.checksum + 1);
   des_ecb_encrypt((des_cblock *)authent.dat, (des_cblock *)authent.dat,
      schedule, 1);
 
   /* Convert response to string and place in buffer */
   p = kdata_to_str((char *)authent.dat, authent.length);

   if (p) {
#ifdef APXS2
      sprintf(errstr, "[%s] User %s authenticated", p, r->user);
#else
      sprintf(errstr, "[%s] User %s authenticated", p, r->connection->user);
#endif /* APXS2 */
      free(p);
      ap_table_set (r->headers_out, "WWW-Authenticate", errstr);
   } else {
      /* Out of memory */
      sprintf(errstr, "Not enough memory to create reply, eek!");
      ap_log_reason (errstr, r->uri, r);
      /* let somebody else take care of it */
      return sec->krb_auth_fail_status;
   }

   return OK;
}
#endif
#ifdef KRB5
static int kerb_validate_ticket_V5( request_rec *r )
{
   kerb_auth_config_rec *sec =
      (kerb_auth_config_rec *)ap_get_module_config(r->per_dir_config,
                              &kerb_auth_module);
   krb5_data         k5authent;
   krb5_error_code   code;
   krb5_context      k5context;
   krb5_keytab       k5keytabid = NULL;
   krb5_auth_context *k5auth_context = NULL;
   krb5_principal    serverp;
   krb5_principal    clientp;
   krb5_ticket       *k5ticket = NULL;
   krb5_data         k5ap_rep_data;
   char *type, *p;
   char errstr[MAX_STRING_LEN], tmpstr[MAX_KDATA_LEN];
   const char *sent_pw, *auth_line;
   char k5_keytab[MAX_STRING_LEN] = "";

   auth_line = (char *) ap_table_get (r->headers_in, "Authorization");

   /* did they send us the information required? */
   if ( auth_line == NULL ) {
      /* tell the client what to expect */
      if(sec->krb_auth_fail_status == HTTP_UNAUTHORIZED)
          ap_table_set (r->err_headers_out, "WWW-Authenticate", "KerberosV5");
      return sec->krb_auth_fail_status;
   }

   /* get the auth type sent to us */
   type = ap_getword(r->pool, &auth_line, ' ');

   /*
    * These two fields in the connection record are used elsewhere
    * in the Apache modules. (check_access)
    * They really appreciate them being set to useful values. :)
    */
#ifdef APXS2
   r->user = ap_getword(r->pool, &auth_line, ' ');
   r->ap_auth_type = "KerberosV5";
#else
   r->connection->user = ap_getword(r->pool, &auth_line, ' ');
   r->connection->ap_auth_type = "KerberosV5";
#endif /* APXS2 */

   sent_pw = ap_getword(r->pool, &auth_line, '\0');

   if ((k5authent.length = str_to_kdata(sent_pw, tmpstr)) == -1) {
      sprintf(errstr,"Invalid Kerberos authenticator");
      /* log it to the error file */
      ap_log_reason (errstr, r->uri, r);
      /* let somebody else take care of it */
      return sec->krb_auth_fail_status;
   }

   k5authent.data = tmpstr;

   code = krb5_init_context(&k5context);

   if (code) {
      sprintf(errstr, "krb5_init_context(): Error - %s", error_message(code));
      ap_log_reason (errstr, r->uri, r);
      return sec->krb_auth_fail_status;
   }

   krb5_init_ets(k5context);

   /* find server principal name; NULL means krb libs determine my hostname.
    * result is khttp/LOCALHOST@LOCALREALM
    */
   code = krb5_sname_to_principal(k5context,
      NULL, "khttp", KRB5_NT_SRV_HST, &serverp);

   if (code) {
      sprintf(errstr,
         "krb5_sname_to_principal(): Error finding server principal name: %s",
         error_message(code));
      ap_log_reason (errstr, r->uri, r);
      return sec->krb_auth_fail_status;
   }

   if (sec->krb5_keytab) {
      strncpy(k5_keytab, sec->krb5_keytab, MAX_STRING_LEN);
   }
#ifdef KRB_V5_KEYTAB
   else {
      strncpy(k5_keytab, KRB_V5_KEYTAB, MAX_STRING_LEN);
   }
#endif /* KRB_V5_KEYTAB */

   /* Check for user-specified keytab */
   if (k5_keytab[0]) {
      code = krb5_kt_resolve(k5context, k5_keytab, &k5keytabid);
      if (code) {
         sprintf(errstr, "krb5_kt_resolve(): Error resolving keytab file: %s",
            error_message(code));
         ap_log_reason (errstr, r->uri, r);
         return sec->krb_auth_fail_status;
      }
   }

   /* and most importantly, check the client's authenticator */
   code = krb5_rd_req(k5context,
      k5auth_context, &k5authent, serverp, k5keytabid, NULL, &k5ticket);

   if (code) {
      sprintf(errstr, "krb5_rd_req(): Error - %s", error_message(code));
      ap_log_reason (errstr, r->uri, r);
      return sec->krb_auth_fail_status;
   }

   clientp = k5ticket->enc_part2->client;
   
   /* send an AP_REP message to complete mutual authentication */
   code = krb5_mk_rep(k5context, *k5auth_context, &k5ap_rep_data);

   if (code) {
      sprintf(errstr, "krb5_mk_rep(): Error - %s", error_message(code));
      ap_log_reason (errstr, r->uri, r);
      return sec->krb_auth_fail_status;
   }

   /* Convert response to string and place in buffer */
   p = kdata_to_str(k5ap_rep_data.data, k5ap_rep_data.length);

   if (p) {
#ifdef APXS2
      sprintf(errstr, "[%s] User %s authenticated", p, r->user);
#else
      sprintf(errstr, "[%s] User %s authenticated", p, r->connection->user);
#endif /* APXS2 */
      free(p);
      ap_table_set (r->headers_out, "WWW-Authenticate", errstr);
   } else {
      /* Out of memory */
      sprintf(errstr, "Not enough memory to create reply, eek!");
      ap_log_reason (errstr, r->uri, r);
      return sec->krb_auth_fail_status;
   }

   return OK;
}
#endif

/*************************************************************************
 * kerb_authenticate_user -- entry point for kerberos authentication
 *
 * Accepts:  request record
 * Returns:  OK, DECLINED, HTTP_UNAUTHORIZED or SERVER_ERROR
 *
 ************************************************************************/
static int kerb_authenticate_user (request_rec *r)
{
   kerb_auth_config_rec *sec =
      (kerb_auth_config_rec *)ap_get_module_config(r->per_dir_config,
                              &kerb_auth_module);
   int ticket = 1;      /* ticket or no, assume yes */
   int res;             /* fcn result code */
   const char *type, *name;   /* auth information */
   const char *auth_line, *sent_pw;
   int KerberosV4 = 0;  /* authtype v4? */
   int KerberosV5 = 0;  /* authtype v5? */
   char *ptr;           /* string place holder */

   /* get the type specified in .htaccess */
   type = ap_auth_type(r);

   /* get the user realm specified in .htaccess */
   name = ap_auth_name(r);

   /* if AuthType is not Kerberos, then don't touch it */
#ifdef KRB4
   if( (type != NULL) && (strncasecmp(type, "KerberosV4", 10) == 0) ) {
      KerberosV4 = 1;
   }
#endif

#ifdef KRB5
   if( (type != NULL) && (strncasecmp(type, "KerberosV5", 10) == 0) ) {
      KerberosV5 = 1;
   }
#endif

   if ( !(KerberosV4 || KerberosV5) ) {
      if (!(sec->krb_auth_authoritative)) {
	 /* let's try anyway */
	 KerberosV5 = 1;
      } else {
	 /* we decline the offer.... */
	 return DECLINED;
      }
   }

#if defined(KRB5_SAVE_CREDENTIALS) || defined(KRB4_SAVE_TICKETS)
   /*
   * Big Hack: Check for a SaveCredentials on end of authtype.
   */
   if ((type!=NULL) && strstr(type, "SaveCredentials")) {
#ifdef KRB5_SAVE_CREDENTIALS
      if (KerberosV5) {
         krb5_save_credentials = 1;
      }
      else {
         krb5_save_credentials = 0;
      }
#endif /* KRB5_SAVE_CREDENTIALS */
#ifdef KRB4_SAVE_TICKETS
      if (KerberosV4) {
         krb4_save_tickets = 1;
      }
      else {
         krb4_save_tickets = 0;
      }
#endif /* KRB4_SAVE_TICKETS */
   }
#endif /* KRB5_SAVE_CREDENTIALS || KRB4_SAVE_TICKETS */

   /* get what the user sent us in the HTTP header */
   auth_line = (char *) ap_table_get (r->headers_in, "Authorization");
    
   /* did we get an auth line? */
   if(!auth_line) {
      /* nope, well, tell the client what to send */

      if(sec->krb_auth_fail_status == HTTP_UNAUTHORIZED)/* KrbAuthFailStatus */
#if (defined(KRB4)) || (defined(KRB5))
      /* check for those browsers that we know need basic anyways */
      /* @JER - this left for backwards compatability */
      if (ap_table_get(r->subprocess_env, "use_basic_auth")) {

         ap_table_set (r->err_headers_out, "WWW-Authenticate",
			"Basic realm=\"Kerberos\"");

      } else if (ap_table_get(r->subprocess_env, "use_kerberos_auth")) {
#ifdef KRB4
         if ( KerberosV4 ) {
            ap_table_set (r->err_headers_out, "WWW-Authenticate",
               ap_pstrcat(r->pool, "KerberosV4 realm=\"", name, "\"", NULL));
         }
#endif

#ifdef KRB5
         if ( KerberosV5 ) {
            ap_table_set (r->err_headers_out, "WWW-Authenticate",
               ap_pstrcat(r->pool, "KerberosV5 realm=\"", name, "\"", NULL));
         }
#endif

      } else {
         /* this is the new "more correct" default behavior */
         ap_table_set (r->err_headers_out, "WWW-Authenticate",
            ap_pstrcat(r->pool, "Basic realm=\"", name, "\"", NULL));
      }
#endif

      return sec->krb_auth_fail_status;
   }
 
   /*
    * Did they send us a Basic auth?
    */
   if (strncasecmp(ap_getword (r->pool, &auth_line, ' '), "Basic", 5) == 0) {
      /*
       * No problem!   Let's get the userid/passwd pair they sent and 
       * see if we can buy them a ticket.
       */
      sent_pw = (char *) ap_pbase64decode (r->pool, auth_line);

      /*
       * These two fields in the connection record are used elsewhere
       * in the Apache modules. (check_access)
       * They really appreciate them being set to useful values. :)
       */
#ifdef APXS2
      r->user = ap_getword (r->pool, &sent_pw, ':');
      r->ap_auth_type = "Basic";
#else
      r->connection->user = ap_getword (r->pool, &sent_pw, ':');
      r->connection->ap_auth_type = "Basic";
#endif /* APXS2 */

      /* flag this as a ticket-less connection for processing */
      ticket = 0;
   }
 
#ifdef KRB4
   /* do this for each connection since it can be set per directory */
   kerb_get_server_realm(r);
#endif

   /* based on what we have so far, let's see if we can help them out */
   if ( ticket ) {
      /* what? a ticket? must have a hacked browser */
#ifdef KRB4
      if ( KerberosV4 ) {
         res = kerb_validate_ticket_V4(r);
      }
#endif

#ifdef KRB5
      if ( KerberosV5 ) {
         res = kerb_validate_ticket_V5(r);
      }
#endif
   } else {

      /* do not allow user to override realm setting of server */
#ifdef APXS2
      if (strchr(r->user,'@')) {
#else
      if (strchr(r->connection->user,'@')) {
#endif /* APXS2 */
	ap_log_reason ("specifying realm in user name is prohibited", r->uri, r);
	return sec->krb_auth_fail_status;
      }

      /* do they know the magic word? */
#ifdef KRB4
      if ( KerberosV4 ) {
         res = kerb_validate_user_pass_V4(r, sent_pw);
#if (defined(DUAL_AUTH)) && (defined(KRB5))
         if ( res != OK ) {
            res = kerb_validate_user_pass_V5(r, sent_pw);
         }
#endif /* DUAL_AUTH && KRB5 */
      }
#endif /* KRB4 */

#ifdef KRB5
      if ( KerberosV5 ) {
         res = kerb_validate_user_pass_V5(r, sent_pw);
#if (defined(DUAL_AUTH)) && (defined(KRB4))
         if ( res != OK ) {
            res = kerb_validate_user_pass_V4(r, sent_pw);
         }
#endif /* DUAL_AUTH && KRB4 */
      }
#endif /* KRB5 */

      memset((char *)sent_pw, 0x00, strlen(sent_pw));
   }

   /*
    * Send a WWW-Authenticate: header back to the client even if they
    * supplied a bad password - as per HTTP standards.
    *
    * The WWW-Authenticate: header must be sent along with every 401
    * response, which returning HTTP_UNAUTHORIZED will generate.
    */

   if ( res == HTTP_UNAUTHORIZED ) {
#if (defined(KRB4)) || (defined(KRB5))
      /* check for those browsers that we know need basic anyways */
      if (ap_table_get(r->subprocess_env, "use_basic_auth")) {
         ap_table_set (r->err_headers_out, "WWW-Authenticate",
            "Basic realm=\"Kerberos\"");
      } else if (ap_table_get(r->subprocess_env, "use_kerberos_auth")) {
#ifdef KRB4
         if ( KerberosV4 ) {
            ap_table_set (r->err_headers_out, "WWW-Authenticate",
               ap_pstrcat(r->pool, "KerberosV4 realm=\"", name, "\"", NULL));
         }
#endif
#ifdef KRB5
         if ( KerberosV5 ) {
            ap_table_set (r->err_headers_out, "WWW-Authenticate",
               ap_pstrcat(r->pool, "KerberosV5 realm=\"", name, "\"", NULL));
         }
#endif
      } else {
         /* this is the new "more correct" default behavior */
         ap_table_set (r->err_headers_out, "WWW-Authenticate",
            ap_pstrcat(r->pool, "Basic realm=\"", name, "\"", NULL));
      }
#endif /* KRB4 || KRB5 */
   }
 
   return res; 
}

#ifdef HESIOD
/*******************************************************************************
 * check_user_access -- check user authentication against hesiod
 *
 * Accepts:  request record
 * Returns:  OK, DECLINED, HTTP_UNAUTHORIZED or SERVER_ERROR
 *
 ******************************************************************************/
int check_user_access (request_rec *r) {
    kerb_auth_config_rec *sec =
      (kerb_auth_config_rec *)ap_get_module_config(r->per_dir_config,
                              &kerb_auth_module);
#ifdef APXS2
    char *user = r->user;
#else
    char *user = r->connection->user;
#endif /* APXS2 */
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
#ifdef APXS2
    const apr_array_header_t *reqs_arr = ap_requires (r);
#else
    const array_header *reqs_arr = ap_requires (r);
#endif /* APXS2 */
    require_line *reqs;
#ifdef APXS2
    apr_table_t *grpstatus;
#else
    table *grpstatus;
#endif /* APXS2 */
    char **user_groups, *grpret;
    char errstr[50];
    int i;

    if (!reqs_arr)
        return OK;
    reqs = (require_line *)reqs_arr->elts;

    for (x=0; x < reqs_arr->nelts; x++) {

      if (! (reqs[x].method_mask & (1 << m))) {
         continue;
      }

      method_restricted = 1;

      t = reqs[x].requirement;  /* t = "require valid-user/user/group u g" */
      w = ap_getword(r->pool, &t, ' ');   /* w = "valid-user/user/group" */

      if (!strcmp(w,"valid-user")) {
           return OK;
      } else if (!strcmp(w,"user")) {
         while (t[0]) {
             w = ap_getword_conf(r->pool, &t);
             if (!strcmp(user, w))
                 return OK;
         }
      } else if (!strcmp(w,"group")) {

#ifdef APXS2
         user_groups = hes_resolve(r->user, "grplist");
#else
         user_groups = hes_resolve(r->connection->user, "grplist");
#endif /* APXS2 */

         if (!user_groups) {
            return sec->krb_auth_fail_status; /* no groups found? */
         }

#ifdef APXS2
         grpstatus = (apr_table_t *)user_groups[0];
#else
         grpstatus = (table *)user_groups[0];
#endif /* APXS2 */

         while (t[0]) {
            w = ap_getword(r->pool, &t, ' '); /* w = "u", then "g" */
            grpret = strstr((char *)grpstatus, w);

            if (grpret) {
               i = strlen(w);
               if ((grpret[i]==':')||(grpret[i]==NULL)) {
                  return OK;
               }
            }
         }
      }
   }

   return sec->krb_auth_fail_status;
}
#endif /* HESIOD */

/*************************************************************************
 * set_fail_status_slot()
 * reads KrbAuthFailStatus directive text into an int
 * sets (int) krb_auth_fail_status to either 401 or 403
 * defaults to 401 under all circumstances (no errors)
 * HTTP_UNAUTHORIZED = (default value) = 401
 * HTTP_FORBIDDEN = (non-negotiable authorisation failure) = 403
 * anilg@geocities.com ngulati@scu.edu.au
 ************************************************************************/
#ifdef APXS2
static const char *set_fail_status_slot(cmd_parms *cmd, void *offset,
				const char *arg)
{
	if (!strncasecmp(arg, "unauthorized", 12))
		return ap_set_int_slot(cmd, offset, "401");
	else if (!strncasecmp(arg, "forbidden", 9))
		return ap_set_int_slot(cmd, offset, "403");
	else
		return apr_pstrcat(cmd->pool, "KrbAuthFailStatus must be Forbidden or Unauthorized", NULL);
	return NULL;
}
#else
static const char *set_fail_status_slot(cmd_parms *cmd, char *struct_ptr,
                                        char *arg)
{
    int offset = (int) (long) cmd->info;
    *(int *) (struct_ptr + offset) = HTTP_UNAUTHORIZED;
    if (!strncasecmp(arg, "unauthorized", 12))
        *(int *) (struct_ptr + offset) = HTTP_UNAUTHORIZED;
    else if (!strncasecmp(arg, "forbidden", 9))
        *(int *) (struct_ptr + offset) = HTTP_FORBIDDEN;
    else return "KrbAuthFailStatus must be Forbidden or Unauthorized";
    return NULL;
}
#endif /* APXS2 */

/*************************************************************************
 * create_kerb_auth_dir_config -- does the configuration
 *
 * Accepts:  memory pool and dummy char
 * Returns:  pointer to allocated config record
 *
 ************************************************************************/
#ifdef APXS2
static void *create_kerb_auth_dir_config (apr_pool_t *p, char *d)
#else
static void *create_kerb_auth_dir_config (pool *p, char *d)
#endif /* APXS2 */
{
    static void *rec;
    rec = (void *) ap_pcalloc (p, sizeof(kerb_auth_config_rec));
    ((kerb_auth_config_rec *)rec)->krb_auth_fail_status = HTTP_UNAUTHORIZED;
    ((kerb_auth_config_rec *)rec)->krb_auth_authoritative = 1; /* secure by default */
    return rec;
}


/* register the functions in the correct places */
#ifdef APXS2
static void register_hooks(apr_pool_t *p)
{
	ap_hook_check_user_id(kerb_authenticate_user, NULL, NULL, APR_HOOK_MIDDLE);
#ifdef HESIOD
	ap_hook_access_checker(check_user_access, NULL, NULL, APR_HOOK_MIDDLE);
#endif /* HESIOD */
}


module AP_MODULE_DECLARE_DATA kerb_auth_module =
{
   STANDARD20_MODULE_STUFF,
   create_kerb_auth_dir_config,   /* dir config creater */
   NULL,                          /* dir merger */
   NULL,                          /* server config */
   NULL,                          /* merge server config */
   kerb_auth_cmds,                /* command apr_table_t */
   register_hooks                 /* register hooks */
};
#else
module kerb_auth_module = {
   STANDARD_MODULE_STUFF,
   NULL,                          /* initializer */
   create_kerb_auth_dir_config,   /* dir config creater */
   NULL,                          /* dir merger --- default is to override */
   NULL,                          /* server config */
   NULL,                          /* merge server config */
   kerb_auth_cmds,                /* command table */
   NULL,                          /* handlers */
   NULL,                          /* filename translation */
   kerb_authenticate_user,        /* check_user_id */
#ifdef HESIOD
   check_user_access,               /* check auth */
#else
   NULL,                          /* check auth */
#endif /* HESIOD */
   NULL,                          /* check access */
   NULL,                          /* type_checker */
   NULL,                          /* fixups */
   NULL                           /* logger */
};
#endif /* APXS2 */
